package cn.skcks.docking.gb28181.wvp.service.video;

import cn.hutool.core.io.IoUtil;
import cn.hutool.core.util.IdUtil;
import jakarta.servlet.ServletOutputStream;
import jakarta.servlet.http.HttpServletResponse;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.bytedeco.ffmpeg.global.avcodec;
import org.bytedeco.ffmpeg.global.avutil;
import org.bytedeco.javacv.*;
import org.springframework.core.io.buffer.DataBuffer;
import org.springframework.core.io.buffer.DefaultDataBuffer;
import org.springframework.core.io.buffer.DefaultDataBufferFactory;
import org.springframework.http.HttpHeaders;
import org.springframework.http.MediaType;
import org.springframework.http.server.reactive.ServerHttpResponse;
import org.springframework.stereotype.Service;
import reactor.core.publisher.Mono;

import java.io.*;
import java.net.http.HttpClient;
import java.nio.file.Path;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

@Slf4j
@Service
public class RecordService {
    public void header(ServerHttpResponse response){
        HttpHeaders headers = response.getHeaders();
        headers.setContentType(MediaType.APPLICATION_OCTET_STREAM);
        headers.set("Accept-Ranges","none");
        headers.setConnection("close");
    }

    // @Async(Default)
    @SneakyThrows
    public Mono<Void> record(ServerHttpResponse response, String url, long timeout){
        Mono<Void> mono = Mono.empty();
        //        response.setContentType(MediaType.APPLICATION_OCTET_STREAM_VALUE);
        header(response);
        Path tmp = Path.of(System.getProperty("java.io.tmpdir"), IdUtil.getSnowflakeNextIdStr()).toAbsolutePath();
        File file = new File(tmp + ".mp4");
        log.info("创建文件 {}, {}", file, file.createNewFile());
        log.info("url {}", url);
        DataBuffer dataBuffer = response.bufferFactory().allocateBuffer(DefaultDataBufferFactory.DEFAULT_INITIAL_CAPACITY);
        OutputStream outputStream = dataBuffer.asOutputStream();
        try (FFmpegFrameGrabber grabber = new FFmpegFrameGrabber(url)) {
            grabber.start();

            try (FFmpegFrameRecorder recorder = new FFmpegFrameRecorder(file, grabber.getImageWidth(), grabber.getImageHeight(), grabber.getAudioChannels())) {
                recorder.start();
                log.info("开始录像");
                log.info("{}", file);
                recorder.setVideoCodec(avcodec.AV_CODEC_ID_H264);
                recorder.setPixelFormat(avutil.AV_PIX_FMT_YUV420P); // 视频源数据yuv
                recorder.setAudioCodec(avcodec.AV_CODEC_ID_AAC); // 设置音频压缩方式
                recorder.setFormat("mp4");
                recorder.setVideoOption("threads", String.valueOf(Runtime.getRuntime().availableProcessors())); // 解码线程数

                ScheduledExecutorService scheduledExecutorService = Executors.newSingleThreadScheduledExecutor();
                AtomicBoolean record = new AtomicBoolean(true);
                scheduledExecutorService.schedule(() -> {
                    log.info("到达超时时间, 结束录制");
                    record.set(false);
                }, timeout, TimeUnit.SECONDS);

                try {
                    Frame frame;
                    while (!response.isCommitted() && record.get() && (frame = grabber.grab()) != null) {
                        recorder.record(frame);
                    }
                    grabber.stop();
                    recorder.stop();
                } catch (FFmpegFrameRecorder.Exception | FrameGrabber.Exception e) {
                    throw new RuntimeException(e);
                }
            }
            // response.getOutputStream()

        } catch (Exception e){
            log.info("{}",e.getMessage());
        }finally {
            log.info("结束录制");
            if(!response.isCommitted()){
                InputStream inputStream = new BufferedInputStream(new FileInputStream(file));
                try{
                    IoUtil.copy(inputStream, outputStream);
                } catch (Exception ignore){}
                log.info("临时文件 {} 写入 响应", file);
                mono = response.writeWith(Mono.just(dataBuffer));
            }
            log.info("删除临时文件 {} {}", file, file.delete());
        }
        return mono;
    }
}
